/*
 *      nlconnections.c
 *      
 *      Copyright 2012 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
 
#include <glib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <netinet/tcp.h>
#include <netinet/in.h>
#include <linux/netlink.h>
#include <linux/inet_diag.h>
#include <dirent.h>
#include <ctype.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "nlconnections.h"


static gboolean nlconnections_numeric_name(gchar *dirname)
{
	if ((dirname == NULL) || (!*dirname)) return FALSE;
	
	while (*dirname) {
		if (!isdigit(*dirname)) {
			return FALSE;
		} else {
			*dirname++;
		}
	}
	
	return TRUE;
}

static gboolean nlconnections_process_access(gchar *dirname, uid_t uid)
{
	gchar fullpath[128];
	struct stat pathstat;
	
	if (!nlconnections_numeric_name(dirname)) return FALSE;
	
	memset(fullpath, 0, sizeof(fullpath));
	
	sprintf(fullpath, "/proc/%s", dirname);
	
	if (stat(fullpath, &pathstat) == -1) {
		return FALSE;
	}
	
	if (pathstat.st_uid != uid) {
		return FALSE;
	}
	
	return TRUE;
}

static gboolean nlconnections_socket_access(gchar *dirname, gchar *sockname, guint inode)
{
	gchar fullpath[128];
	struct stat fdstat;
	
	if (!nlconnections_numeric_name(sockname)) return FALSE;
		
	memset(fullpath, 0, sizeof(fullpath));
	
	sprintf(fullpath, "/proc/%s/fd/%s", dirname, sockname);
	
	if (stat(fullpath, &fdstat) == -1) {
		return FALSE;
	}
	
	if (((fdstat.st_mode & S_IFMT) == S_IFSOCK) && (fdstat.st_ino == inode)) {
		return TRUE;
	}
	
	return FALSE;
}

static gchar *nlconnections_process_name(gchar *dirname, gchar *appname, gsize appsize)
{
	gint fd, i;
	gchar fpath[128];
	ssize_t linkchars;
		
	if ((dirname == NULL) || (dirname[0] == '\0')) return NULL;
	if ((appname == NULL) || (appsize == 0)) return NULL;
	
	memset(fpath, 0, sizeof(fpath));
	sprintf(fpath, "/proc/%s/exe", dirname);
	
	linkchars = readlink(fpath, appname, appsize-1);
	
	if (linkchars == 0) {
		memset(fpath, 0, sizeof(fpath));
		sprintf(fpath, "/proc/%s/comm", dirname);
		
		fd = open(fpath, O_RDONLY);
		if (fd != -1) {
			linkchars = read(fd, appname, appsize-1);
			close(fd);
		} else {
			return NULL;
		}
	}
		
	appname[linkchars] = '\0';
	
	for (i=linkchars; i>=0; i--) {
		if (appname[i] == '/') {
			memmove(appname+0, appname+i+1, linkchars-i-1);
			linkchars -= i+1;
			break;
		}
	}
	
	appname[linkchars] = '\0';
	
	return appname;
}

static gboolean nlconnections_get_process(guint inode, gchar *appname, gsize namesize, pid_t *apppid)
{
	DIR *procdir, *fddir;
	struct dirent *procde, *fdde;
	char fdirpath[128], fdpath[128];
	
	if ((appname == NULL) || (namesize == 0) || (apppid == NULL)) return FALSE;
	
	procdir = opendir("/proc");
	if (procdir != NULL) {
		while (procde = readdir(procdir)) {
			if (nlconnections_process_access(procde->d_name, getuid())) {
				memset(fdirpath, 0, sizeof(fdirpath));
				sprintf(fdirpath, "/proc/%s/fd", procde->d_name);
				//enumerate file descriptors
				fddir = opendir(fdirpath);
				if (fddir != NULL) {
					while (fdde = readdir(fddir)) {
						if (nlconnections_socket_access(procde->d_name, fdde->d_name, inode)) {
							//printf("%s:%s (%s)\n", procde->d_name, fdde->d_name, nlconnections_process_name(procde->d_name, appname, sizeof(appname)));
							*apppid = atoi(procde->d_name);
							nlconnections_process_name(procde->d_name, appname, namesize);
							closedir(fddir);
							closedir(procdir);
							return TRUE;
						}
					}
					closedir(fddir);
				}
			}
		}
		closedir(procdir);
	}
	
	return FALSE;
}

gboolean nlconnections_terminate_application(pid_t pid)
{
	if (kill(pid, 0) == 0) {
		if (kill(pid, SIGTERM) == 0) {
			return TRUE;
		}
	}
	
	return FALSE;
}

gchar *nlconnections_socket_state(guchar state)
{
	switch (state) {
		case TCP_ESTABLISHED:
			return "Established";
		case TCP_SYN_SENT:
			return "SYN sent";
		case TCP_SYN_RECV:
			return "SYN recv";
		case TCP_FIN_WAIT1:
			return "FIN wait";
		case TCP_FIN_WAIT2:
			return "FIN wait";
		case TCP_TIME_WAIT:
			return "Time wait";
		case TCP_CLOSE:
			return "Close";
		case TCP_CLOSE_WAIT:
			return "Close wait";
		case TCP_LAST_ACK:
			return "Last ACK";
		case TCP_LISTEN:
			return "Listen";
		case TCP_CLOSING:
			return "Closing";
		default:
			return "Unknown";
	}
}

void nlconnections_hash_destroy(gpointer data)
{
	nlconnections_socket_t socket;
	
	socket = (nlconnections_socket_t)data;
	
	if (socket == NULL) return;
	
	if (socket->appname != NULL) g_free(socket->appname);
	if (socket->dsthostname != NULL) g_free(socket->dsthostname);
	
	g_free(socket);
}

static gboolean nlconnections_hash_clear_foreach(gpointer key, gpointer value, gpointer user_data)
{
	nlconnections_socket_t socket;
	time_t currenttime;
	
	socket = (nlconnections_socket_t)value;
	currenttime = *(time_t *)user_data;
	
	if (socket->updatetime == currenttime) {
		return FALSE;
	} else {
		//printf("Remove: %u\n", socket->inode);
		return TRUE;
	}
}

gboolean nlconnections_update(nlconnections_data_t data)
{
	gchar msgbuf[8192];
	struct iovec iov;
	struct msghdr message;
	struct nlmsghdr *header;
	struct nlmsgerr *error;
	struct inet_diag_msg *entry;
	struct hostent *dsthost;
	gchar srcbuf[INET6_ADDRSTRLEN];
	gchar dstbuf[INET6_ADDRSTRLEN];
	gchar appname[1024];
	pid_t apppid;
	time_t currenttime;
	gint res;
	nlconnections_socket_t socket;
	gboolean updateres;
	
	if (data == NULL) return FALSE;
	
	if (sendmsg(data->nlfd, &data->message, 0) < 0) {
		return FALSE;
	}
	
	memset(&msgbuf, 0, sizeof(msgbuf));
	iov.iov_base = msgbuf;
	iov.iov_len = sizeof(msgbuf);
	
	currenttime = time(NULL);
	
	while (1) {
		memset(&message, 0, sizeof(struct msghdr));
		message.msg_name = (void *)&data->nladdr;
		message.msg_namelen = sizeof(struct sockaddr_nl);
		message.msg_iov = &iov;
		message.msg_iovlen = 1;
		
		res = recvmsg(data->nlfd, &message, 0);
		
		if (res <= 0) {
			if ((res < 0) && (errno == EINTR)) {
				continue;
			} else {
				updateres = FALSE;
				goto clear;
			}
		}
		
		header = (struct nlmsghdr *)msgbuf;
		
		while (NLMSG_OK(header, res)) {
			if (header->nlmsg_type == NLMSG_DONE) {
				//printf("NLMSG_DONE\n");
				updateres = TRUE;
				goto clear;
			}
			if (header->nlmsg_type == NLMSG_ERROR) {
				error = (struct nlmsgerr*)NLMSG_DATA(header);
				//printf("NLMSG_ERROR %s\n", strerror(-(error->error)));
				updateres = FALSE;
				goto clear;
			}
			
			entry = (struct inet_diag_msg *)NLMSG_DATA(header);
			
			if (entry != NULL) {
				if ((entry->idiag_uid == data->userid) || (data->userid == 0)) {
					if (!g_hash_table_contains(data->connections, (gconstpointer)&entry->idiag_inode)) {
						if (nlconnections_get_process(entry->idiag_inode, appname, sizeof(appname), &apppid)) {
							socket = g_new(struct _nlconnections_socket, 1);
							socket->inode = entry->idiag_inode;
							socket->family = entry->idiag_family;
							socket->userid = entry->idiag_uid;
							socket->updatetime = currenttime;
							socket->dqueue = entry->idiag_rqueue + entry->idiag_wqueue;
							socket->state = entry->idiag_state;
							socket->srcport = ntohs(entry->id.idiag_sport);
							g_snprintf(socket->srcaddr, sizeof(socket->srcaddr), "%s:%u", inet_ntop(entry->idiag_family, entry->id.idiag_src, srcbuf, INET6_ADDRSTRLEN), ntohs(entry->id.idiag_sport));
							g_snprintf(socket->dstaddr, sizeof(socket->dstaddr), "%s:%u", inet_ntop(entry->idiag_family, entry->id.idiag_dst, dstbuf, INET6_ADDRSTRLEN), ntohs(entry->id.idiag_dport));
							socket->appname = g_strdup(appname);
							socket->apppid = apppid;
							socket->dsthostname = NULL;
							dsthost = gethostbyaddr(entry->id.idiag_dst, sizeof(entry->id.idiag_dst), entry->idiag_family);
							if (dsthost != NULL) {
								socket->dsthostname = g_strdup(dsthost->h_name);
							} else {
								socket->dsthostname = g_strdup(socket->dstaddr);
							}
							g_hash_table_insert(data->connections, (gpointer)&socket->inode, socket);
							//printf("Insert: %u\n", entry->idiag_inode);
						}
					} else {
						socket = g_hash_table_lookup(data->connections, (gconstpointer)&entry->idiag_inode);
						if (socket != NULL) {
							socket->updatetime = currenttime;
							socket->dqueue = entry->idiag_rqueue + entry->idiag_wqueue;
							socket->state = entry->idiag_state;
							//printf("Update: %u\n", entry->idiag_inode);
						}					
					}
				}
			}
			
			header = NLMSG_NEXT(header, res);
		}
	}
	
	clear:
	
	g_hash_table_foreach_remove(data->connections, nlconnections_hash_clear_foreach, &currenttime);
		
	return updateres;
}

void nlconnections_close(nlconnections_data_t data)
{
	if (data == NULL) return;
	
	close(data->nlfd);
	
	g_hash_table_destroy(data->connections);
	
	g_free(data);
}

nlconnections_data_t nlconnections_open(void)
{
	nlconnections_data_t data;
	
	data = g_new(struct _nlconnections_data, 1);
	
	#ifndef NETLINK_SOCK_DIAG
		data->nlfd = socket(AF_NETLINK, SOCK_RAW, NETLINK_INET_DIAG);
	#else
		data->nlfd = socket(AF_NETLINK, SOCK_RAW, NETLINK_SOCK_DIAG);
	#endif
	
	if (data->nlfd < 0) {
		g_free(data);
		return NULL;
	}
	
	memset(&data->request.nlhdr, 0, sizeof(struct nlmsghdr));
	data->request.nlhdr.nlmsg_len = sizeof(struct socket_request);
	data->request.nlhdr.nlmsg_type = TCPDIAG_GETSOCK;
	data->request.nlhdr.nlmsg_flags = NLM_F_REQUEST | NLM_F_ROOT;
	data->request.nlhdr.nlmsg_pid = 0;
	data->request.nlhdr.nlmsg_seq = 0;
	
	memset(&data->request.nlreq, 0, sizeof(struct inet_diag_req));
	data->request.nlreq.idiag_family = AF_INET | AF_INET6;
	data->request.nlreq.idiag_states = ((1 << TCP_CLOSING + 1) - 1);
	
	data->reqiov.iov_base = &data->request;
	data->reqiov.iov_len = sizeof(struct socket_request);
	
	memset(&data->nladdr, 0, sizeof(struct sockaddr_nl));
	data->nladdr.nl_family = AF_NETLINK;
	data->nladdr.nl_pid = 0;
	data->nladdr.nl_groups = 0;
	
	memset(&data->message, 0, sizeof(struct msghdr));
	data->message.msg_name = (void *)&data->nladdr;
	data->message.msg_namelen = sizeof(struct sockaddr_nl);
	data->message.msg_iov = &data->reqiov;
	data->message.msg_iovlen = 1;
	
	data->userid = getuid();
	
	data->connections = g_hash_table_new_full(g_int_hash, g_int_equal, NULL, (GDestroyNotify)nlconnections_hash_destroy);
	
	return data;
}
