/*
 *      smsdb.c
 *      
 *      Copyright 2012 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdlib.h>
#include <glib.h>
#include <glib/gprintf.h>
#include <string.h>
#include <gdbm.h>
#include <ctype.h>

//#include "dbus-api.h"
#include "encoding.h"
#include "smsdb.h"

#define SMSDB_SMS_MESSAGE_XML "<sms>\n\t<number>%s</number>\n\t<time>%lu</time>\n\t<text>%s</text>\n\t<read>%u</read>\n\t<folder>%u</folder>\n</sms>\n\n"

#define SMSDB_READ_TAG        "\n\t<read>"
#define SMSDB_TRAILER_TAG     "\n</sms>\n\n"
#define SMSDB_TRAILER_PARAMS  "\n\t<read>%u</read>\n\t<folder>%u</folder>\n</sms>\n\n"

enum _gmm_smsdb_xml_elements {
	SMSDB_XML_PARAM_NUMBER = 0,
	SMSDB_XML_PARAM_TIME,
	SMSDB_XML_PARAM_TEXT,
	SMSDB_XML_PARAM_READ,
	SMSDB_XML_PARAM_FOLDER,
	SMSDB_XML_PARAM_NULL
};

static gint gmm_smsdb_xml_parameter = SMSDB_XML_PARAM_NULL;

static gint gmm_smsdb_sms_message_sort_compare(gconstpointer a, gconstpointer b);
static void gmm_smsdb_free_sms_list_foreach(gpointer data, gpointer user_data);
static sms_full_message_t gmm_smsdb_xml_parse(gchar *xml, gsize size);
static void gmm_smsdb_xml_get_element(GMarkupParseContext *context, const gchar *element, const gchar **attr_names, const gchar **attr_values, gpointer data, GError **error);
static void gmm_smsdb_xml_get_value(GMarkupParseContext *context, const gchar *text, gsize size, gpointer data, GError **error);
static void gmm_smsdb_xml_end_element(GMarkupParseContext *context, const gchar *element, gpointer data, GError **error);


smsdb_t gmm_smsdb_open(gchar *appname, gchar *deviceid)
{
	smsdb_t smsdb;
	const gchar *homepath; 
	gchar filename[64];
	gchar *confpath;
	
	if ((appname == NULL) || (deviceid == NULL)) return NULL;
	
	homepath = g_get_home_dir();
	
	if (homepath == NULL) return NULL;
	
	confpath = g_build_filename(homepath, ".config", appname, NULL);
	
	if (g_mkdir_with_parents(confpath, 0755) != 0) {
		g_warning("Cant create program settings directory");
		g_free(confpath);
		return NULL;
	}
	
	g_free(confpath);
	
	smsdb = g_new(struct _smsdb, 1);
	
	g_snprintf(filename, sizeof(filename), "sms-%s.gdbm", deviceid);
	
	smsdb->filepath = g_build_filename(homepath, ".config", appname, filename, NULL);
	
	return smsdb;
}

gboolean gmm_smsdb_close(smsdb_t smsdb)
{
	if (smsdb == NULL) return FALSE;
	
	if (smsdb->filepath != NULL) {
		g_free(smsdb->filepath);
	}
	
	g_free(smsdb);
	
	return TRUE;
}

void smsdb_free_sms(sms_full_message_t sms)
{
	if (sms == NULL) return;
	
	if (sms->number != NULL) {
		g_free(sms->number);
		sms->number = NULL;
	}
	
	if (sms->idents != NULL) {
		g_array_free(sms->idents, TRUE);
		sms->idents = NULL;
	}
	
	if (sms->message != NULL) {
		g_string_free(sms->message, TRUE);
		sms->message = NULL;
	}
	
	g_free(sms);
}

gboolean gmm_smsdb_add_sms(smsdb_t smsdb, sms_full_message_t sms)
{
	GDBM_FILE db;
	gchar smsid[64];
	gulong idvalue;
	gint idlen;
	datum key, data;
	gchar *smsnumber;
	gchar *smstext;
	gchar *smsxml;
	
		
	if ((smsdb == NULL) || (sms == NULL)) return FALSE;
	if (smsdb->filepath == NULL) return FALSE;
	if ((sms->number == NULL) || (sms->message->str == NULL)) return FALSE;
		
	db = gdbm_open(smsdb->filepath, 0, GDBM_WRCREAT, 0755, 0);
	
	if (db == NULL) return FALSE;
	
	do {
		idvalue = (gulong)random();
		memset(smsid, 0, sizeof(smsid));
		idlen = snprintf(smsid, sizeof(smsid), "%lu", idvalue);
		key.dptr = (gchar *)smsid;
		key.dsize = idlen;
	} while (gdbm_exists(db, key));
	
	sms->dbid = idvalue;
	
	smsnumber = g_markup_escape_text(sms->number, -1);
	
	if (smsnumber == NULL) {
		g_warning("Unable to convert SMS number string");
		gdbm_close(db);
		return FALSE;
	}
	
	smstext = g_markup_escape_text(sms->message->str, -1);
	
	if (smstext == NULL) {
		g_warning("Unable to convert SMS text string");
		g_free(smsnumber);
		gdbm_close(db);
		return FALSE;
	}
	
	smsxml = g_strdup_printf(SMSDB_SMS_MESSAGE_XML, smsnumber, sms->timestamp, smstext, sms->read, sms->folder);
		
	data.dptr = smsxml;
	data.dsize = strlen(smsxml);
	
	if (gdbm_store(db, key, data, GDBM_REPLACE) == -1) {
		g_warning("Unable to write to database");
		gdbm_close(db);
		g_free(smsxml);
		return FALSE;
	}
	
	gdbm_sync(db);
	gdbm_close(db);
	
	g_free(smsxml);
	g_free(smsnumber);
	g_free(smstext);
	
	return TRUE;
}

static gint gmm_smsdb_sms_message_sort_compare(gconstpointer a, gconstpointer b)
{
	sms_full_message_t sms1, sms2;
		
	sms1 = (sms_full_message_t)a;
	sms2 = (sms_full_message_t)b;
	
	if (sms1->timestamp < sms2->timestamp) {
		return -1;
	} else if (sms1->timestamp > sms2->timestamp) {
		return 1;
	} else {
		return 0;
	}
}

GSList *gmm_smsdb_read_sms_list(smsdb_t smsdb)
{
	GDBM_FILE db;
	GSList *list;
	sms_full_message_t sms;
	datum key, data;
	gchar smsid[64];
	
	if (smsdb == NULL) return NULL;
	if (smsdb->filepath == NULL) return NULL;
	
	db = gdbm_open(smsdb->filepath, 0, GDBM_READER, 0755, 0);
	
	if (db == NULL) return NULL;
	
	list = NULL;
	
	key = gdbm_firstkey(db);
	
	if (key.dptr != NULL) {
		do {
			data = gdbm_fetch(db, key);
			if (data.dptr != NULL) {
				sms = gmm_smsdb_xml_parse(data.dptr, data.dsize);
				if (sms != NULL) {
					memset(smsid, 0, sizeof(smsid));
					strncpy(smsid, key.dptr, key.dsize);
					sms->dbid = strtoul(smsid, NULL, 0);
					list = g_slist_prepend(list, sms);
				}
			}
			key = gdbm_nextkey(db, key);
		} while (key.dptr != NULL);
	}
	
	gdbm_close(db);
	
	if (list != NULL) {
		list = g_slist_sort(list, gmm_smsdb_sms_message_sort_compare);
	} 
	
	return list;
}

static void gmm_smsdb_free_sms_list_foreach(gpointer data, gpointer user_data)
{
	sms_full_message_t sms;
	
	if (data != NULL) return;
	
	sms = (sms_full_message_t)data;
	
	if (sms->number != NULL) {
		g_free(sms->number);
		sms->number = NULL;
	}
	
	if (sms->idents != NULL) {
		g_array_free(sms->idents, TRUE);
		sms->idents = NULL;
	}
	
	if (sms->message != NULL) {
		g_string_free(sms->message, TRUE);
		sms->message = NULL;
	}
}

void gmm_smsdb_free_sms_list(GSList *smslist)
{
	if (smslist == NULL) return;
	
	g_slist_foreach(smslist, gmm_smsdb_free_sms_list_foreach, NULL);
	g_slist_free(smslist);
}

sms_full_message_t gmm_smsdb_read_sms_message(smsdb_t smsdb, gulong idvalue)
{
	GDBM_FILE db;
	gchar smsid[64];
	gint idlen;
	datum key, data;
	sms_full_message_t sms;
		
	if (smsdb == NULL) return NULL;
	if (smsdb->filepath == NULL) return NULL;
	
	db = gdbm_open(smsdb->filepath, 0, GDBM_READER, 0755, 0);
	
	if (db == NULL) return NULL;
	
	sms = NULL;
	
	memset(smsid, 0, sizeof(smsid));
	idlen = snprintf(smsid, sizeof(smsid), "%lu", idvalue);
	key.dptr = (gchar *)smsid;
	key.dsize = idlen;
	
	if (gdbm_exists(db, key)) {
		data = gdbm_fetch(db, key);
		if (data.dptr != NULL) {
			sms = gmm_smsdb_xml_parse(data.dptr, data.dsize);
			sms->dbid = idvalue;
		}
	}
	
	gdbm_close(db);
	
	return sms;
}

gboolean gmm_smsdb_remove_sms_message(smsdb_t smsdb, gulong idvalue)
{
	GDBM_FILE db;
	gchar smsid[64];
	gint idlen;
	datum key, data;
	
	if (smsdb == NULL) return FALSE;
	if (smsdb->filepath == NULL) return FALSE;
	
	db = gdbm_open(smsdb->filepath, 0, GDBM_WRCREAT, 0755, 0);
	
	if (db == NULL) return FALSE;
	
	memset(smsid, 0, sizeof(smsid));
	idlen = g_snprintf(smsid, sizeof(smsid), "%lu", idvalue);
	key.dptr = (gchar *)smsid;
	key.dsize = idlen;
	
	if (gdbm_exists(db, key)) {
		if (gdbm_delete(db, key) == 0) {
			gdbm_sync(db);
			gdbm_close(db);
			return TRUE;
		}
	}
	
	gdbm_close(db);
	
	return FALSE;
}

gboolean gmm_smsdb_set_message_read_status(smsdb_t smsdb, gulong idvalue, gboolean readflag)
{
	GDBM_FILE db;
	gchar smsid[64];
	gint idlen;
	datum key, data;
	gchar *node, *trailer, *newmsg;
	gchar newtrailer[64];
	gint newtrailerlen;
	gboolean res;
			
	if (smsdb == NULL) return FALSE;
	if (smsdb->filepath == NULL) return FALSE;
	
	db = gdbm_open(smsdb->filepath, 0, GDBM_WRITER, 0755, 0);
	
	if (db == NULL) return FALSE;
	
	memset(smsid, 0, sizeof(smsid));
	idlen = snprintf(smsid, sizeof(smsid), "%lu", idvalue);
	key.dptr = (gchar *)smsid;
	key.dsize = idlen;
	
	res = FALSE;
	
	if (gdbm_exists(db, key)) {
		data = gdbm_fetch(db, key);
		if (data.dptr != NULL) {
			node = strstr(data.dptr, SMSDB_READ_TAG);
			if (node != NULL) {
				if ((node-data.dptr > 8) && (isdigit(node[8]))) {
					if (readflag) {
						node[8] = '1';
					} else {
						node[8] = '0';
					}
					
					if (gdbm_store(db, key, data, GDBM_REPLACE) == 0) {
						res = TRUE;
					}
					
					free(data.dptr);
				}
			} else {
				if (strstr(data.dptr, SMSDB_TRAILER_TAG) != NULL) {
					memset(newtrailer, 0, sizeof(newtrailer));
					newtrailerlen = g_snprintf(newtrailer, sizeof(newtrailer), SMSDB_TRAILER_PARAMS, readflag, SMSDB_SMS_FOLDER_INCOMING);
					
					newmsg = g_malloc0(data.dsize-9+newtrailerlen+1);
					memcpy(newmsg, data.dptr, data.dsize-9);
					memcpy(newmsg+data.dsize-9, newtrailer, newtrailerlen);
					
					free(data.dptr);
					
					data.dptr = newmsg;
					data.dsize = data.dsize-9+newtrailerlen;
					
					if (gdbm_store(db, key, data, GDBM_REPLACE) == 0) {
						res = TRUE;
					}
					
					g_free(newmsg);
				}
			}
		}
	}
	
	gdbm_close(db);
	
	return res;
}

static sms_full_message_t gmm_smsdb_xml_parse(gchar *xml, gsize size)
{
	sms_full_message_t sms;
	GMarkupParser mp;
	GMarkupParseContext *mpc;
	GError *error = NULL;
	
	sms = g_new(struct _sms_full_message, 1);
	
	sms->read = FALSE;
	sms->folder = SMSDB_SMS_FOLDER_INCOMING;
	sms->idents = NULL;
	
	mp.start_element = gmm_smsdb_xml_get_element;
	mp.end_element = gmm_smsdb_xml_end_element;
	mp.text = gmm_smsdb_xml_get_value;
	mp.passthrough = NULL;
	mp.error = NULL;
	
	mpc = g_markup_parse_context_new(&mp, 0, (gpointer)sms, NULL);
	g_markup_parse_context_parse(mpc, xml, size, &error);
	if (error != NULL) {
		//g_warning(error->message);
		g_error_free(error);
		g_markup_parse_context_free(mpc);
		return NULL;
	}
	g_markup_parse_context_free(mpc);
	
	return sms;
}

static void gmm_smsdb_xml_get_element(GMarkupParseContext *context, const gchar *element, const gchar **attr_names, const gchar **attr_values, gpointer data, GError **error)
{
	if (g_str_equal(element, "number")) gmm_smsdb_xml_parameter = SMSDB_XML_PARAM_NUMBER;
	else if (g_str_equal(element, "time")) gmm_smsdb_xml_parameter = SMSDB_XML_PARAM_TIME;
	else if (g_str_equal(element, "text")) gmm_smsdb_xml_parameter = SMSDB_XML_PARAM_TEXT;
	else if (g_str_equal(element, "read")) gmm_smsdb_xml_parameter = SMSDB_XML_PARAM_READ;
	else if (g_str_equal(element, "folder")) gmm_smsdb_xml_parameter = SMSDB_XML_PARAM_FOLDER;
}

static void gmm_smsdb_xml_get_value(GMarkupParseContext *context, const gchar *text, gsize size, gpointer data, GError **error)
{
	sms_full_message_t sms;
	gchar *numescstr, *textescstr;
	gsize numlen;
	
	sms = (sms_full_message_t)data;
	
	if (gmm_smsdb_xml_parameter == SMSDB_XML_PARAM_NULL) return;
	
	switch (gmm_smsdb_xml_parameter) {
		case SMSDB_XML_PARAM_NUMBER:
			numescstr = encoding_unescape_xml_markup(text, size);
			if (numescstr != NULL) {
				sms->number = bcd_to_utf8_ascii_part(numescstr, strlen(numescstr), &numlen);
				g_free(numescstr);
			} else {
				sms->number = bcd_to_utf8_ascii_part(text, size, &numlen);
			}
			if (sms->number != NULL) {
				sms->number = encoding_clear_special_symbols(sms->number, numlen);
			} else {
				sms->number = encoding_clear_special_symbols(g_strdup(text), size);
			}
			break;
		case SMSDB_XML_PARAM_TIME:
			sms->timestamp = (time_t)atol(text);
			break;
		case SMSDB_XML_PARAM_TEXT:
			textescstr = encoding_unescape_xml_markup(text, size);
			if (textescstr != NULL) {
				sms->message = g_string_new(textescstr);
				g_free(textescstr);
			} else {
				sms->message = g_string_new(text);
			}
			break;
		case SMSDB_XML_PARAM_READ:
			if (atoi(text)) {
				sms->read = TRUE;
			} else {
				sms->read = FALSE;
			}
			break;
		case SMSDB_XML_PARAM_FOLDER:
			sms->folder = atoi(text);
			break;
		default:
			break;
	}
}

static void gmm_smsdb_xml_end_element(GMarkupParseContext *context, const gchar *element, gpointer data, GError **error)
{
	if (!g_str_equal(element, "sms")) {
		gmm_smsdb_xml_parameter = SMSDB_XML_PARAM_NULL;
	}
}
